;;; nxml-enc.el --- XML encoding auto-detection

;; Copyright (C) 2003 Thai Open Source Software Center Ltd

;; Author: James Clark <jjc@thaiopensource.com>
;; Keywords: XML

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of
;; the License, or (at your option) any later version.

;; This program is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY; without even the implied
;; warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
;; PURPOSE.  See the GNU General Public License for more details.

;; You should have received a copy of the GNU General Public
;; License along with this program; if not, write to the Free
;; Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
;; MA 02111-1307 USA

;;; Commentary:

;; User entry points are nxml-start-auto-coding and nxml-stop-auto-coding.
;; This is separate from nxml-mode, because this cannot be autoloaded.
;; It may use `xmltok-get-declared-encoding' which can be autoloaded.
;; It's separate from rng-auto.el so it can be byte-compiled, and
;; because it provides independent, useful functionality.

;;; Code:

(defvar nxml-file-name-ignore-case
  (memq system-type '(vax-vms windows-nt)))

(defvar nxml-cached-file-name-auto-coding-regexp nil)
(defvar nxml-cached-auto-mode-alist nil)

(defun nxml-file-name-auto-coding-regexp ()
  "Return regexp for filenames for which XML auto-coding should be done."
  (if (eq auto-mode-alist nxml-cached-auto-mode-alist)
      nxml-cached-file-name-auto-coding-regexp
    (let ((alist auto-mode-alist)
	  (case-fold-search nxml-file-name-ignore-case)
	  regexps)
      (setq nxml-cached-auto-mode-alist alist)
      (while alist
	(when (eq (cdar alist) 'nxml-mode)
	  (setq regexps (cons (caar alist) regexps)))
	(setq alist (cdr alist)))
      (setq nxml-cached-file-name-auto-coding-regexp
	    (if (null (cdr regexps))
		(car regexps)
	      (mapconcat (lambda (r)
			   (concat "\\(?:" r "\\)"))
			 regexps
			 "\\|"))))))

(defvar nxml-non-xml-set-auto-coding-function nil
  "The function that `set-auto-coding-function' should call for non-XML files.")
(defun nxml-set-auto-coding (file-name size)
  (if (let ((case-fold-search nxml-file-name-ignore-case)
	    (regexp (nxml-file-name-auto-coding-regexp)))
	(and regexp
	     (string-match regexp file-name)))
      (nxml-detect-coding-system (+ (point) (min size 1024)))
    (and nxml-non-xml-set-auto-coding-function
	 (funcall nxml-non-xml-set-auto-coding-function file-name size))))

(defun nxml-detect-coding-system (limit)
  (if (< limit (+ (point) 2))
      'utf-8
    (let ((first-two-chars (list (char-after)
				 (char-after (1+ (point))))))
      (cond ((equal first-two-chars '(#xFE #xFF))
	     (and (coding-system-p 'utf-16-be) 'utf-16-be))
	    ((equal first-two-chars '(#xFF #xFE))
	     (and (coding-system-p 'utf-16-le) 'utf-16-le))
	    ((equal first-two-chars '(?< ?\?))
	     (let ((enc (xmltok-get-declared-encoding limit)))
	       (cond (enc
		      (setq enc (intern (downcase enc)))
		      (if (and (coding-system-p enc)
			       ;; Using random Emacs coding system names
			       ;; in an XML encoding declaration
			       ;; is not correct.
			       (eq (coding-system-get enc 'mime-charset) enc))
			  enc
			;; what to return?
			;; raw-text might not work with UTF-16/32 stuff
			'no-conversion))
		     (t 'utf-8))))
	    ;; There are other things we might try here
	    ;; eg UTF-8 BOM, UTF-16 with no BOM 
	    ;; translate to EBCDIC
	    (t 'utf-8)))))

(defun nxml-start-auto-coding ()
  "Do encoding auto-detection as specified in the XML standard.
Applied to any files that `auto-mode-alist' says should be handled by
`nxml-mode'."
  (interactive)
  (unless (eq set-auto-coding-function 'nxml-set-auto-coding)
    (let ((inhibit-quit t))
      (setq nxml-non-xml-set-auto-coding-function set-auto-coding-function)
      (setq set-auto-coding-function 'nxml-set-auto-coding))))

(defun nxml-stop-auto-coding ()
  "Stop doing encoding auto-detection as specified in the XML standard."
  (interactive)
  (when (eq set-auto-coding-function 'nxml-set-auto-coding)
    (let ((inhibit-quit t))
      (setq set-auto-coding-function nxml-non-xml-set-auto-coding-function)
      (setq nxml-non-xml-set-auto-coding-function nil))))

(unless (coding-system-p 'us-ascii)
  (make-coding-system
   ;; Unicode Emacs uses ?- last time I looked
   'us-ascii 2 ?-
   "ISO 2022 based 7-bit encoding for ASCII (MIME:US-ASCII)"
   '(ascii)
   '((safe-charsets ascii)
     (mime-charset . us-ascii))))

;; Work around bug in Emacs 21.3

(when (and (coding-system-p 'utf-16-le)
	   (eq (coding-system-get 'utf-16-le 'pre-write-conversion)
	       'utf-16-le-pre-write-conversion))
  (coding-system-put 'utf-16-le 'pre-write-conversion nil))

(when (and (coding-system-p 'utf-16-le)
	   (eq (coding-system-get 'utf-16-be 'pre-write-conversion)
	       'utf-16-be-pre-write-conversion))
  (coding-system-put 'utf-16-be 'pre-write-conversion nil))

(provide 'nxml-enc)

;;; nxml-enc.el ends here
